@@ -10,7 +10,7 @@ import tempfile
 import configparser
 from subprocess import run, PIPE, DEVNULL
 
-version = "0.9.1"
+version = "0.9.3"
 
 
 class PassNode(urwid.AttrMap):
@@ -91,19 +91,20 @@ class PassList(urwid.ListBox):
         list_navigation_offsets = {
             'down': 1,
             'up': -1,
+            # overshoot to go to bottom/top
             'end': len(self.body),
             'home': -len(self.body),
             'down_screen': size[1],
             'up_screen': -size[1],
             'down_half_screen': size[1] // 2,
             'up_half_screen': -size[1] // 2,
-            # overshoot to go to bottom/top
         }
 
         dir_navigation_directions = {
-            'confirm': 'down',  # the confirm key doubles as enter folder key
+            # the confirm key doubles as enter folder key
+            'confirm': 'down',
             'dir_down': 'down',
-            'dir_up':   'up',
+            'dir_up': 'up',
         }
 
         action = config.keybindings.get(key)
@@ -136,38 +137,46 @@ class PassList(urwid.ListBox):
         offset = self.get_focus_offset_inset(self._size)[0]
 
         # either specify a shift offset, or an absolute position
-        if new_focus is None:
-            new_focus = self.focus_position + shift
-        else:
+        if new_focus:
             shift = new_focus - self.focus_position
+        else:
+            new_focus = shift + self.focus_position
         new_offset = offset + shift
 
         # border check
-        new_focus = min(max(new_focus, 0), len(self.body) - 1)
-        new_offset = min(max(new_offset, 0), self._size[1] - 1)
+        new_focus = min(max(0, new_focus), len(self.body) - 1)
+        new_offset = min(max(0, new_offset), self._size[1] - 1)
 
         self.change_focus(self._size, new_focus, offset_inset=new_offset)
         self._ui.update_view()
 
     def insert(self, node):
         def insert_relative(r, n):
-            n1, sep, n2 = n.partition(os.sep)
-            if sep == os.sep:
+            # if starts with /, then assume the node is relative to store root
+            if n.startswith('/'):
+                n = n.lstrip('/')
+                r = ''
+
+            # separate at the first /
+            n1, sep, n2 = n.partition('/')
+            # recursively insert if there are more levels
+            if sep == '/':
                 insert_relative(os.path.join(r, n1), n2)
-            passnode = PassNode(n1, r, isdir=(sep == os.sep))
+
             # change stored list
+            passnode = PassNode(n1, r, sep == '/')
             if Pass.all_pass.get(r) is None:
                 Pass.all_pass[r] = FolderWalker(r)
-            pos = Pass.all_pass[r].insert_sorted(passnode)
-            # change saved cursor position
-            Pass.all_pass[r].pos = pos
-            return pos
+            Pass.all_pass[r].pos = Pass.all_pass[r].insert_sorted(passnode)
 
-        inserted_pos = insert_relative(self.root, node)
+            # do not change cursor position if the path is not relative
+            return Pass.all_pass[r].pos if r == self.root else None
+
+        inserted_pos = insert_relative(self.root, node.strip())
         # change listwalker
         self.body[:] = Pass.all_pass[self.root]
         # focus the new node
-        self.set_focus(inserted_pos)
+        self.list_navigate(new_focus=inserted_pos)
 
         self._ui.update_view()
 
@@ -179,6 +188,7 @@ class PassList(urwid.ListBox):
 
         self._ui.update_view()
 
+    # TODO: this seems odd being here
     def update_root_count(self):
         for n in Pass.all_pass[os.path.dirname(self.root)]:
             if n.node == self.root and n.isdir:
@@ -190,15 +200,17 @@ class FolderWalker(list):
     def __init__(self, root, dirs=[], files=[]):
         self.pos = 0  # cursor position
 
-        self[:] = [PassNode(f, root, True) for f in sorted(dirs)] + \
-            [PassNode(f, root) for f in sorted(files)]
+        self[:] = [PassNode(f, root, True) for f in sorted(dirs, key=str.lower)] + \
+                  [PassNode(f, root) for f in sorted(files, key=str.lower)]
 
         # prevent empty list, which troubles listbox operations
         if len(self) == 0:
-            self[:] = [PassNode(None, None)]
+            super().append(PassNode(None, None))
 
     def pop(self, index=-1):
         super().pop(index)
+
+        # add a empty placeholder
         if len(self) == 0:
             super().append(PassNode(None, None))
 
@@ -214,13 +226,18 @@ class FolderWalker(list):
 
         # insert and sort, with directories sorted before files
         super().insert(self.pos, node)
-        self[:] = sorted([n for n in self if n.isdir], key=lambda n: n.node) + \
-            sorted([n for n in self if not n.isdir], key=lambda n: n.node)
+        self[:] = sorted([n for n in self if n.isdir], key=lambda n: n.node.lower()) + \
+            sorted([n for n in self if not n.isdir], key=lambda n: n.node.lower())
         return self.index(node)
 
 
+# TODO: multiline insert, this should be easy since we have the workaround in Pass.edit
+# TODO: mv, cp support
+# TODO: QR code generate, maybe?
 # TODO: background preview, or/and cache preview results
-# TODO: deal with focus changing when editing
+# TODO: CLI arguments
+# TODO: git support
+# TODO: otp support
 class UI(urwid.Frame):
     def __init__(self):
         self._last_preview = None
@@ -232,9 +249,14 @@ class UI(urwid.Frame):
         self._help_string = ' a:generate e:edit i:insert y:copy z:toggle /:search'
 
         # widgets
+        self.app_string = urwid.Text(('border', '{}: '.format(self._app_string)))
         self.path_indicator = urwid.Text('', wrap='clip')
         self.help_text = urwid.Text(self._help_string)
-        self.header_widget = urwid.Columns([self.path_indicator, ('pack', self.help_text)])
+        self.header_widget = urwid.Columns([
+            ('pack', self.app_string),
+            urwid.AttrMap(self.path_indicator, 'bright'),
+            ('pack', self.help_text)
+        ])
         self.messagebox = urwid.Text('')
         self.count_indicator = urwid.Text('', align='right')
         self.footer_widget = urwid.Columns([
@@ -276,24 +298,32 @@ class UI(urwid.Frame):
             self.middle.contents = [(self.listbox, ('weight', 1, False))]
         self.middle.focus_position = 0
 
+    def mouse_event(self, size, event, button, col, row, focus):
+        logging.debug(f"ui mouse event: {size} {event} {button} {col} {row} {focus}")
+        # Prevent focus change due to clicking when editing
+        r = self.contents['footer'][0].rows(size[:1], True)
+        if self._edit_type is None or self._edit_type and row >= size[1] - r:
+            super().mouse_event(size, event, button, col, row, focus)
+
     def keypress(self, size, key):
         logging.debug("ui keypress: {} {}".format(key, size))
         action = config.keybindings.get(key)
         if action == 'cancel':
             self.unfocus_edit()
-        elif action == 'quit' and self._edit_type is None:
-            raise urwid.ExitMainLoop
         elif self._edit_type == "copy":
             self.unfocus_edit()
             self.copy_by_key(key)
         elif self._edit_type == "delete":
             self.unfocus_edit()
             self.delete_confirm(key)
-        elif action == 'confirm' and self._edit_type is not None:
-            self.handle_input()
         elif self._edit_type is not None:
-            # pass through to edit widget (the focused widget)
-            return super().keypress(size, key)
+            if action == 'confirm':
+                self.handle_input()
+            else:
+                # pass through to edit widget (the focused widget)
+                return super().keypress(size, key)
+        elif action == 'quit':
+            raise urwid.ExitMainLoop
         elif action == 'search' or action == 'search_back':
             self.focus_edit("search", '/' if action == 'search' else '?')
             self._search_direction = 1 if action == 'search' else -1
@@ -305,7 +335,7 @@ class UI(urwid.Frame):
         elif action == 'generate':
             self.focus_edit("generate", 'Generate a password file: ')
         elif action == 'edit' and not self.listbox.focus.isdir:
-            self.run_pass(Pass.edit, self.listbox.insert,
+            self.run_pass(Pass.edit, None,
                           self.listbox.focus.node, self.listbox.root, "Edit: {}")
             urwid.emit_signal(self, 'redraw')
         elif action == 'delete':
@@ -318,6 +348,8 @@ class UI(urwid.Frame):
         elif action == 'toggle_preview':
             self._preview_shown = not self._preview_shown
             self.update_preview_layout()
+        elif action == 'find':
+            self.focus_edit("find", "Find: ")
         else:
             return super().keypress(size, key)
 
@@ -328,7 +360,7 @@ class UI(urwid.Frame):
         self.messagebox.set_text('')
         self.editbox.set_mask(None)
 
-    def focus_edit(self, edit_type, cap='', mask=None):
+    def focus_edit(self, edit_type, cap, mask=None):
         self._edit_type = edit_type
         self.contents['footer'] = (self.editbox, None)
         self.set_focus('footer')
@@ -337,23 +369,22 @@ class UI(urwid.Frame):
         self.editbox.set_edit_text('')
 
     def handle_input(self):
-        if self._edit_type == "search":
+        edit_type = self._edit_type
+        self.unfocus_edit()
+        if edit_type == "search":
             self._search_pattern = self.editbox.edit_text
-            self.unfocus_edit()
             self.search_in_dir(self._search_pattern, 1)
-        elif self._edit_type == "generate":
-            self.unfocus_edit()
+        elif edit_type == "generate":
             self.run_pass(Pass.generate, self.listbox.insert,
                           self.editbox.edit_text, self.listbox.root, "Generate: {}")
             self.listbox.update_root_count()
-        elif self._edit_type == "insert":
+        elif edit_type == "insert":
             self._insert_node = self.editbox.edit_text
-            self.focus_edit("insert_password", 'Enter password: ', '*')
-        elif self._edit_type == "insert_password":
+            self.focus_edit("insert_password", 'Enter password: ', mask='*')
+        elif edit_type == "insert_password":
             self._insert_pass = self.editbox.edit_text
-            self.focus_edit("insert_password_confirm", 'Enter password again: ', '*')
-        elif self._edit_type == "insert_password_confirm":
-            self.unfocus_edit()
+            self.focus_edit("insert_password_confirm", 'Enter password again: ', mask='*')
+        elif edit_type == "insert_password_confirm":
             self._insert_pass_again = self.editbox.edit_text
             if self._insert_pass == self._insert_pass_again:
                 self.run_pass(Pass.insert, self.listbox.insert,
@@ -362,13 +393,12 @@ class UI(urwid.Frame):
                 self.listbox.update_root_count()
             else:
                 self.message("Password is not the same", alert=True)
+        elif edit_type == "find":
+            self.message("searching " + self.editbox.edit_text)
 
     def update_view(self):
         # update header
-        self.path_indicator.set_text([
-            ('border', '{}: '.format(self._app_string)),
-            ('bright', '/{}'.format(self.listbox.root)),
-        ])
+        self.path_indicator.set_text('/{}'.format(self.listbox.root))
 
         # update footer
         self.count_indicator.set_text("{}/{}".format(
@@ -398,24 +428,30 @@ class UI(urwid.Frame):
             preview = res.stderr if res.returncode else res.stdout
         self.preview.original_widget.set_text(preview)
 
-    def run_pass(self, func, lfunc, node, root, msg, args=(), largs=()):
+    def run_pass(self, func, lfunc, node, root, msg='', args=(), largs=()):
+        # do not accept password name ends with /, pass itself has problems
+        if node.endswith('/'):
+            self.message(f'Can not create a directory: {node}.', alert=True)
+            return
+
         path = os.path.join(root, node)
         res = func(path, *args)
         if res.returncode == 0:
-            self.message(msg.format(path))
-            if lfunc == self.listbox.insert:
-                lfunc(node)
-            elif lfunc == self.listbox.delete:
-                lfunc(largs[0])
-            self.update_preview(force=True)
+            self.message(msg.format(path) if func == Pass.show else '')
+            if lfunc:
+                lfunc(node if lfunc == self.listbox.insert else largs[0])
+            self.update_preview(True)
         else:
             self.message(res.stderr, alert=True)
 
+        return res
+
     def delete_confirm(self, key):
         if key in ['y', 'Y', 'd', 'enter']:
             self.run_pass(Pass.delete, self.listbox.delete,
                           self.listbox.focus.node, self.listbox.root,
                           "Deleting {}", largs=(self.listbox.focus_position,))
+            # TODO: put this into lower level functions
             self.listbox.update_root_count()
         elif key in ['n', 'N']:
             self.message("Abort.")
@@ -424,15 +460,14 @@ class UI(urwid.Frame):
 
     def parse_pass(self, passwd):
         lines = passwd.split('\n')
-        copiable_fields = {str(i + 1): lines[i] for i in range(min(9, len(lines)))}
-        copiable_fields['a'] = passwd
-        copiable_fields['y'] = lines[0]
+        copiable_fields = {'a': passwd, 'y': lines[0], '1': lines[0]}
 
-        for line in lines[1:]:
-            if line.find(':') != -1:
-                field, value = line.split(':')
-                if field in config.copy_bindings:
-                    copiable_fields[config.copy_bindings[field]] = value.strip()
+        for i in range(1, len(lines)):
+            field, sep, value = [s.strip() for s in lines[i].partition(':')]
+            if i < 10:
+                copiable_fields[str(i + 1)] = value if sep == ':' else field
+            if sep == ':' and field in config.copy_bindings:
+                copiable_fields[config.copy_bindings[field]] = value
 
         return copiable_fields
 
@@ -447,10 +482,11 @@ class UI(urwid.Frame):
             if res.returncode != 0:
                 self.message(res.stderr, alert=True)
                 return
+            self.run_pass(Pass.show, None, self.listbox.focus.node, self.listbox.root)
             password = res.stdout
 
-        pw = self.parse_pass(password.strip('\n'))
-        self.focus_edit("copy", 'Copy [{}]: '.format(''.join(pw.keys())))
+        pw = self.parse_pass(password.rstrip('\n'))
+        self.focus_edit("copy", 'Copy [{}]: '.format(''.join(sorted(pw))))
         self._parsed_password = pw
 
     def copy_by_key(self, key):
@@ -467,7 +503,7 @@ class UI(urwid.Frame):
 
     def search_in_dir(self, pattern, direction):
         """ direction = 1 or -1 to specify the search direction """
-        if self._search_pattern is None:
+        if pattern is None:
             self.message("No search pattern", alert=True)
             return
 
@@ -481,9 +517,10 @@ class UI(urwid.Frame):
             start += 1
         search_list = list(range(start, length)) + list(range(start))
 
+        icase = pattern == pattern.lower()
         for i in search_list[::direction]:
             node = self.listbox.body[i].node
-            if node.find(pattern) != -1:
+            if all([s in (node.lower() if icase else node) for s in pattern.split()]):
                 self.listbox.list_navigate(new_focus=i)
                 return
 
@@ -496,6 +533,11 @@ class Pass:
     X_SELECTION = os.getenv("PASSWORD_STORE_X_SELECTION", "clipboard")
     EDITOR = os.getenv("EDITOR", "vi")
     all_pass = dict()
+    # exit if pass dir does not exit
+    if not os.path.exists(PASS_DIR):
+        print("'{}' or $PASSWORD_STORE_DIR does not exist".format(FALLBACK_PASS_DIR))
+        print("See `man pass` for how to set password storage directory.")
+        exit()
 
     @classmethod
     def extract_all(cls):
@@ -511,8 +553,7 @@ class Pass:
     @staticmethod
     def show(path):
         logging.debug("Showing password for {}".format(path))
-        result = run(['pass', 'show', path], stdout=PIPE, stderr=PIPE, text=True)
-        return result
+        return run(['pass', 'show', path], stdout=PIPE, stderr=PIPE, text=True)
 
     @classmethod
     def edit(cls, path):
@@ -529,30 +570,26 @@ class Pass:
                 return res
             fp.seek(0)
             password = fp.read()
-            res = run(['pass', 'insert', '-m', '-f', path], input=password,
-                      stderr=PIPE, stdout=PIPE)
-            return res
+            return run(['pass', 'insert', '-m', '-f', path], input=password,
+                       stderr=PIPE, stdout=PIPE)
 
     @staticmethod
     def insert(path, password):
         pw = password + '\n' + password + '\n'
-        result = run(['pass', 'insert', '-f', path], input=pw,
-                     stdout=PIPE, stderr=PIPE, text=True)
-        return result
+        return run(['pass', 'insert', '-f', path], input=pw,
+                   stdout=PIPE, stderr=PIPE, text=True)
 
     @staticmethod
     def generate(path):
         command = ['pass', 'generate', '-f', path]
         if config.no_symbols:
             command.append('-n')
-        result = run(command, stdout=PIPE, stderr=PIPE, text=True)
-        return result
+        return run(command, stdout=PIPE, stderr=PIPE, text=True)
 
     @staticmethod
     def delete(path):
         command = ['pass', 'rm', '-r', '-f', path]
-        result = run(command, stdout=PIPE, stderr=PIPE, text=True)
-        return result
+        return run(command, stdout=PIPE, stderr=PIPE, text=True)
 
 
 class MyConfigParser(configparser.RawConfigParser):
@@ -598,6 +635,7 @@ class MyConfigParser(configparser.RawConfigParser):
             'search_back': ['?'],
             'search_next': ['n'],
             'search_prev': ['N'],
+            'find': ['f'],
             'insert': ['i'],
             'generate': ['a'],
             'edit': ['e'],
@@ -626,7 +664,7 @@ class MyConfigParser(configparser.RawConfigParser):
             ('dir',         'light blue',   'default'),
             ('alert',       'light red',    'default'),
             ('bright',      'white',        'default'),
-            ('focus',       'black',        'white'),
+            ('focus',       'standout',     'default'),
             ('focusdir',    'black',        'light blue',   'bold'),
         ]
         # update from configuration file
